import matplotlib.pyplot as plt
import seaborn as sns


def poisson():
    a = []
    f = open("local_poisson_n1000_l2.txt")
    for line in f.readlines():
        if line.strip()!="":
            a.append(float(line))
    f.close()
    a.append(8)
    a.append(10)
    ax = sns.countplot(a, color="dodgerblue")
    sns.lineplot(x=[0,0.25,0.5,1,1.5,2,3,4,5,6,7,8,9,10], y=[1300,2250,2500,2650,2800,2750,1750,1000,450,150,50,20, 15, 5], ax=ax, color="r")
    ax.set_yticks([])
    ax.set_ylabel("")
    ax.set_xticks([])
    ax.set_xlabel("")
    plt.show()


def normal():

    a = []
    f = open("local_normal_n10000.txt")
    for line in f.readlines():
        if line.strip()!="":
            a.append(float(line))
    f.close()
    ax = sns.distplot(a, bins=15, kde=True, rug=False, hist=True, color="r", hist_kws={"color": "dodgerblue",
                                                                                       'edgecolor': 'black',
                                                                                       "linewidth": 2})
    ax.set_yticks([])
    ax.set_ylabel("")
    ax.set_xticks([])
    ax.set_xlabel("")
    plt.show()


def noise():
    a = []
    f = open("local_normal_n1000_noisy.txt")
    for line in f.readlines():
        if line.strip()!="":
            a.append(float(line))
    f.close()
    ax = sns.distplot(a, kde=True, rug=False, hist=True, color="r", hist_kws={"color": "dodgerblue",
                                                                                       'edgecolor': 'black',
                                                                                       "linewidth": 2})
    ax.set_yticks([])
    ax.set_ylabel("")
    ax.set_xticks([])
    ax.set_xlabel("")
    plt.show()


def noise_scatter():
    anscombe = sns.load_dataset("anscombe")
    ax = sns.lmplot(x="x", y="y", data=anscombe.query("dataset == 'III'"),
               ci=None, scatter_kws={"s": 80});
    plt.show()



from tadaqq.slabel import SLabel
from collections import Counter
from tadaqq.util import uri_to_fname, compute_scores


class SLabMer:

    SLAB_PREF = 'slabel_pref'
    CLUS_PREF = 'clus_pref'
    PREFS = [SLAB_PREF, CLUS_PREF]

    def __init__(self, endpoint):
        self.sl = SLabel(endpoint)

    def _assign_preds_and_get_freq(self, group, remove_outliers, estimate, err_meth, k=3):
        """
        Perform semantic labelling for each element get the frequency of predictions for each cluster
        Return: Frequency
        """
        pred_classes = []
        for ele in group:
            self.sl.collect_numeric_data(class_uri=ele['class_uri'])
            pred = self.sl.annotate_column(ele['col'], class_uri=ele['class_uri'], remove_outliers=remove_outliers,
                                           estimate=estimate, err_meth=err_meth)
            pred = pred[:k]
            ele['preds'] = [perr[1] for perr in pred]
            ele['candidates'] = []
            for p in ele['preds']:
                pred_classes.append(p)

        c = Counter(pred_classes)
        freqs = c.most_common()
        return freqs

    def annotate_cluster(self, group, remove_outliers, estimate, err_meth, candidate_failback, pref=None, k=3):
        """
        Store the predicted properties and assign a single candidate for each element in the given group
        :param group:
        :param remove_outliers:
        :param estimate:
        :param err_meth:
        :param k:
        :return:
        """
        if pref not in self.PREFS:
            print("Invalid preference")
            return None

        # Perform semantic labelling for each element get the frequency of predictions for each cluster
        freqs = self._assign_preds_and_get_freq(group=group, remove_outliers=remove_outliers, estimate=estimate,
                                                err_meth=err_meth, k=k)
        # print("\n\nfreqs: ")
        # print(freqs)
        # if freqs == []:
        #     print(group)
        if pref == self.SLAB_PREF:
            return self.annotate_cluster_slabel_pref(group=group, freqs=freqs, candidate_failback=candidate_failback)
        elif pref == self.CLUS_PREF:
            return self.annotate_cluster_clus_pref(group=group, freqs=freqs)

    def annotate_cluster_clus_pref(self, group, freqs):
        """
        Store the predicted properties and assign a single candidate for each element in the given group

        :param group:
        :param freqs: a list of the pairs (property uri, number of occurrences)
        :return:
        """
        # print("\n\nGroup: ")
        for ele in group:
            # if len(ele['candidates']) > 0:
            #     print("More: %s " % str(ele))
            for prop_uri, fre in freqs:
                ele['candidates'].append(prop_uri)
            # print("\tCandidates: %s" % str(ele['candidates']))
            # print("\tclass uri: %s" % ele['class_uri'])
            # print("\tcol id: %d" % ele['col_id'])
            # print("\tfname: %s" % ele['fname'])
            # print("\tproperty: %s " % ele['property'])
            # print("\tcandidate: %s " % str(ele['candidates'][:3]))
            # print("===============================")

    def annotate_cluster_slabel_pref(self, group, freqs, candidate_failback):
        """
        Store the predicted properties and assign a single candidate for each element in the given group
        :param group:
        :param remove_outliers:
        :param estimate:
        :param err_meth:
        :param k:
        :return:
        """

        # print("\n\nGroup: ")
        for ele in group:
            for prop_uri, fre in freqs:
                if prop_uri in ele['preds']:
                    ele['candidates'].append(prop_uri)

            # print("\tclass uri: %s" % ele['class_uri'])
            # print("\tcol id: %d" % ele['col_id'])
            # print("\tfname: %s" % ele['fname'])
            # print("\tproperty: %s " % ele['property'])
            # print("\tcandidate: %s " % str(ele['candidates'][:3]))
            # print("===============================")

            if candidate_failback:
                for p in ele['preds']:
                    if p not in ele['candidates']:
                        ele['candidates'].append(p)

    def annotate_clusters(self, groups, remove_outliers, estimate, err_meth, candidate_failback, pref=SLAB_PREF, k=3):
        for g in groups:
            self.annotate_cluster(group=g, remove_outliers=remove_outliers, estimate=estimate, err_meth=err_meth,
                                  candidate_failback=candidate_failback, k=k, pref=pref)

    def evaluate_labelling(self, groups):
        """
        :param groups: list of groups. Each group is an ele
        :return:
        """
        eval_data = []
        debug_list = []
        debug_k = []
        taget_gid = 4
        for gid, group in enumerate(groups):
            # print("\n\nGroup %d\n========" % gid)
            for ele in group:
                # corr_trans_uris = [uri_to_fname(ele['property'])]
                if gid == taget_gid:
                    debug_list.append(ele['property'])
                    # print("\n\nproperty: %s" % ele['property'])
                    # print("properties: %s" % (str(ele['properties'])))
                    # print("candidate: %s" % ele['candidates'][0])
                    # print("\n\nCandidates: %s" % (str(ele)))
                corr_trans_uris = [uri_to_fname(p) for p in ele['properties']]
                p_errs = [(0.0, cand) for cand in ele['candidates']]
                res = self.sl.eval_column(p_errs, correct_uris=corr_trans_uris, print_diff=False)

                # debug_k.append(res)
                # if gid == taget_gid:
                #     # print("k: %d" % res)
                #     debug_k.append(res)
                eval_data.append(res)
                # print("\t%s" % res)
            # if gid == taget_gid:
                # print("\n")
                # print(Counter(debug_list).most_common())
                # print(Counter(debug_k).most_common())
        # print(Counter(eval_data).most_common())
        prec, rec, f1 = compute_scores(eval_data, k=1)
        score = {'prec': prec, 'rec': rec, 'f1': f1}
        return score





from collections import Counter
from tadaqq.qq import QQE


class Clusterer:

    def __init__(self, save_memory=True):
        self.groups = []
        self.save_memory = save_memory

    def column_group_matching(self, ele, fetch_method, err_meth, err_cutoff, same_class):
        """
            Add column to one of the groups (or to a new group)
        """
        groups = self.groups
        min_idx = None
        min_err = 1
        qq = QQE(list(ele["col"]))
        # print("\n\n groups: %d" % len(self.groups))
        for idx, g in enumerate(self.groups):
            if not g:
                print("group None")
                raise Exception("group is None")
            # The column to be compared is always placed at index 0. This is done in the add_col_to_group method
            top_ele = g[0]
            if top_ele is None:
                raise Exception("top_ele is None")
            err = qq.predict_and_get_error(top_ele["col"], method=err_meth, remove_outliers=False)
            # print("err %f\t%s %s" % (err, ele['property'], top_ele['property']))
            if err < min_err:
                if same_class and top_ele["concept"] != ele["concept"]:
                    # print("class skip")
                    continue
                min_idx = idx
                min_err = err

        if min_err < err_cutoff:
            # print("append cluster")
            group = self.add_col_to_group(ele, groups[min_idx], fetch_method)
            groups[min_idx] = group
        else:
            # print("new cluster")
            # print("min %f" % min_err)
            groups.append([ele])
        if groups[-1] is None:
            print(ele)
            raise Exception("column_group_matching> Exception: no groups is added")

    def add_col_to_group(self, ele, group, fetch_method):
        """
        Add ele to group.
        """
        if fetch_method == "max":
            if group[0]["num"] < ele["num"]:
                group.append(group[0])
                if self.save_memory:
                    group[-1]["col"] = []
                group[0] = ele
            else:
                if self.save_memory:
                    ele["col"] = []
                group.append(ele)
        elif fetch_method == "min":
            if group[0]["num"] > ele["num"]:
                group.append(group[0])
                group[-1]["col"] = []
                group[0] = ele
            else:
                if self.save_memory:
                    ele["col"] = []
                group.append(ele)
        else:
            raise Exception("add_col_to_group> Exception: unknown fetch method")
        return group

    def evaluate(self, counts, print_eval=False):
        """
        groups: [[{}], [{}]]
        counts: Counter of cluster values. Each value = "idx-concept-shot_property"
        """
        if print_eval:
            print("\n\nEVALUATE\n==============\n")
        max_per_v = dict()
        groups = self.groups
        for idx, g in enumerate(groups):
            vals = [ele['gs_clus'] for ele in g]
            c = Counter(vals)
            for k in c:
                prec = c[k] / len(vals)
                if k in max_per_v:
                    if (c[k] > max_per_v[k]['num']) or (c[k] == max_per_v[k]['num'] and prec > max_per_v[k]['prec']):
                        max_per_v[k] = {
                            'num': c[k],
                            'prec': prec,
                            'clus_id': idx
                        }
                else:
                    max_per_v[k] = {
                        'num': c[k],
                        'prec': prec,
                        'clus_id': idx
                    }

        scores = dict()
        precs = []
        recs = []
        f1s = []
        if print_eval:
            print("{:<35} {:<5} {:<5} {:<5} {:<5}".format("name", "prec", "rec", "f1", "clus"))
        for k in max_per_v:
            prec = max_per_v[k]['prec']
            rec = max_per_v[k]['num'] / counts[k]
            f1 = 2 * prec * rec / (prec + rec)
            scores[k] = {'prec': prec, 'rec': rec, 'f1': f1}
            precs.append(prec)
            recs.append(rec)
            f1s.append(f1)
            if print_eval:
                print("{:<35} {:<5} {:<5} {:<5} {:<5}".format(k, round(prec, 3), round(rec, 3), round(f1, 3),
                                                              max_per_v[k]['clus_id']))
        p, r, f = sum(precs) / len(precs), sum(recs) / len(recs), sum(f1s) / len(f1s)
        if print_eval:
            print("Average: Precision (%.3f), Recall (%.3f), F1 (%.3f)" % (p, r, f))
        return p, r, f





import numpy as np


def get_stats(data):
    mean = np.mean(data)
    median = np.median(data)
    std = np.std(data)
    return mean, median, std


def compare_stats(data1, data2):
    """
    :param data1:
    :param data2:
    :return:
    """
    mean1, median1, std1 = get_stats(data1)
    mean2, median2, std2 = get_stats(data2)
    print("%10s     %10s      %10s" % ("Mean", "Median", "STD"))
    print("%10f.2   %10f.2    %10f.2" % (mean1, median1, std1))
    print("%10f.2   %10f.2    %10f.2" % (mean2, median2, std2))




import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

description = """Quantile Quantile Plot with Linear Approximation, Clustering, and Semantic Labelling of Numeric Columns
"""

# The text of the README file
with open("README.md") as f:
    README = f.read()
    lines = README.split('\n')
    desc_lines = [line for line in lines if line[:2] != "[!"]
    README = "\n".join(desc_lines)
    README += "\n"
# This call to setup() does all the work
setup(
    name="tada-qq",
    version="2.3.0",
    description=description,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/oeg-upm/tada-qq",
    author="Ahmad Alobaid",
    author_email="aalobaid@fi.upm.es",
    license="Apache2",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=["tadaqq.qq", "tadaqq.slabel", "tadaqq.util", 'tadaqq.slabmer', 'tadaqq.clus'],
    include_package_data=True,
    install_requires=["pandas", "easysparql", "pcake"]
)
